#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <unistd.h>
#include <arpa/inet.h>

#define MAX_CONNS    16

int Service(int sock);


int main(int argc, char **argv) {
  struct sockaddr_in server;     // struct for server address
  struct sockaddr_in client;     // struct for client address
  int port;                      // server port number
  int listen_sock;               // listening socket
  int newsock;                   // socket for incoming connection
  int connsocks[MAX_CONNS];      // array of connected sockets, for select
  int numconns;                  // number of connections
  char buf[1024];                // buffer for transmit and receive
  int rc;                        // result code
  int asize, i;


  printf("Echo server (multi-client single-process version)\n"
         "Written by Igor Vanin (vanin@gpmail.spb.ru)\n\n");

  // Create a TCP socket to accept incoming connections
  listen_sock = socket(AF_INET, SOCK_STREAM, 0);
  if (listen_sock == -1) {
    perror("tcp socket()");
    exit(1);
  }

  printf("TCP socket assigned is %d\n", listen_sock);

  if (argc >= 2) {
    // Get the port number from command line
    port = atoi(argv[1]);
    if (port > 0 && port < 65535) {
      printf("Using port number %d\n", port);
      port = htons(port);  // change byte order - host to network
    }
    else {
      printf("Incorrect command line. Port number expected\n");
    }
  }
  
  if (argc < 2 || port < 1 || port > 65534) {
    // Get the port for the echo protocol out of the etc/services file
    struct servent *echoprot;
    echoprot = getservbyname("echo", "tcp");
    if (echoprot == NULL) {
      printf("The echo/tcp protocol is not listed in the etc/services file\n");
      exit(1);
    }
    port = echoprot->s_port;
    printf("Using default port number %d\n", ntohs(port));
  }

  server.sin_family = AF_INET;
  server.sin_port = port;
  server.sin_addr.s_addr = INADDR_ANY;

  // Bind the TCP socket to the echo port
  rc = bind(listen_sock, (struct sockaddr *)&server, sizeof(server));
  if (rc == -1) {
    perror("tcp bind()");
    exit(1);
  }

  // Put the socket into listen mode
  rc = listen(listen_sock, SOMAXCONN);
  if (rc == -1) {
    perror("listen()");
    exit(1);
  }

  numconns = 0; // No connections yet

  // Wait for data to echo or a connection on the TCP socket
  while (1) {
    // Setup the socket array for select()
    fd_set r;
    FD_ZERO(&r);
    FD_SET(listen_sock, &r);
    for (i = 0; i < numconns; i++) {
      FD_SET(connsocks[i], &r);
    }

    printf("Waiting for connection%s\n", numconns > 0 ? " or incoming data" : "");
    // Wait indefinitely
    rc = select(FD_SETSIZE, &r, NULL, NULL, NULL);
    if (rc == -1) {
      // If the select was interrupted, just continue
      if (errno == EINTR) continue;
      perror("select()");
      exit(1);
    }

    // Check the TCP sockets to see if we got any data on them
    for (i = 0; i < numconns; i++) {
      if (FD_ISSET(connsocks[i], &r)) {
        // We got some data - handle it
        rc = Service(connsocks[i]);
        if (rc == -1) {
          // The connection was closed - move the last connected socket
          // into the position this socket occupies
          connsocks[i] = connsocks[numconns-1];
          numconns--;
        }
      }
    }

    // If this was a request for a new TCP connection, accept it
    if (FD_ISSET(listen_sock, &r)) {
      // accept() the new connection
      asize = sizeof(client);
      newsock = accept(listen_sock, (struct sockaddr *)&client, &asize);
      if (newsock == -1) {
        perror("accept()");
        exit(1);
      }

      printf("Received a TCP connection on socket %d from %s port %d\n", newsock, inet_ntoa(client.sin_addr), ntohs(client.sin_port));

      // If there are more concurrent connections than we can handle, just
      // close the socket. Doing the accept() and then a close() takes
      // the connection request off our queue, and keeps the client from
      // hanging around forever trying to connect.
      if (numconns >= MAX_CONNS)
        close(newsock);
      else
        connsocks[numconns++] = newsock;
    }
  }
}


// A client sent us some data - echo it back to him
int Service(int sock) {
  int rc, len, actual_len;
  int socks[1];
  char buf[1024];

  // Find out how much data he sent us
  rc = ioctl(sock, FIONREAD, (char *)&actual_len, sizeof(actual_len));
  if (rc == -1) {
    perror("ioctl()");
    exit(1);
  }
  len = actual_len < sizeof(buf) ? actual_len : sizeof(buf);

  // Read the data
  rc = recv(sock, buf, len, 0);
  if (rc == -1) {
    perror("recv()");
    exit(1);
  }

  // The client broke the connection, if recv returned 0
  if (rc == 0) {
    printf("TCP socket %d: The client broke the connection\n", sock);
    close(sock);
    return -1;
  }

  printf("TCP socket %d: received %d bytes\n", sock, len);

  // Echo the data back
  rc = send(sock, buf, len, 0);
  if (rc == -1) {
    perror("send()");
    exit(1);
  }

  return 0;
}
